org 0x7C00  ; set up start address of bootloader

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; setup a stack and segment regs ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

    xor ax, ax
    mov ds, ax
    mov es, ax
    mov ax, 0x7000   ; stack address 
    mov ss, ax       
    xor sp, sp       ; set stackpointer to 0 
	
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;   read kernel from floppy disk                           ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

read_kernel: 
    mov si,loadmsg     
    call print_string
    mov [bootdrive], dl ; boot drive stored by BIOS in DL.

    ; di: number of sectors, bx: segment, ch, dh, cl: cylinder, head, sector
    mov bx, 100
	mov di, bx 
	mov bx, 0x1000 ; kernel start address (cf. linker script)
	mov ch, 0
	mov dh, 0
	mov cl, 2
    call read_sectors
						
;;;;;;;;;;;;;
; A20-Gate  ;
;;;;;;;;;;;;;	

switch_a20:
    in  al, 0x92      ; switch A20 gate via fast A20 port 92
    cmp al, 0xff      ; if it reads 0xFF, nothing's implemented on this port
    je .no_fast_A20
   
    or  al, 2         ; set A20_Gate_Bit (bit 1)
    and al, ~1        ; clear INIT_NOW bit (don't reset pc...)
    out 0x92, al
    jmp .A20_done
   
.no_fast_A20:         ; no fast shortcut -> use the slow kbc...
    call empty_8042 
   
    mov al, 0xD1      ; kbc command: write to output port
    out 0x64, al
    call empty_8042
   
    mov al, 0xDF      ; writing this to kbc output port enables A20
    out 0x60, al
    call empty_8042

.A20_done:

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Load GDT, switch to PM, and jump to kernel ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
		
load_gdt:	
    cli               ; clear interrupts
    lgdt [gdtr]       ; load GDT via GDTR (defined in file "gtd.inc")	

switch_to_PM:	
    mov eax, cr0      ; switch-over to Protected Mode
    or  eax, 1        ; set bit 0 of CR0 register
    mov cr0, eax      ;	

jmp_to_PM:
    jmp dword 08:0x00010000    ; this is a 16-bit-FAR-Jmp, but CS is loaded with "index" 8 in GDT 
                      ; hence, the code will be interpreted as 32 bit from here on
					  ; the address can be found in the linker script 
;;;;;;;;;
; Calls ;
;;;;;;;;;
 
print_string:
    mov ah, 0x0E      ; VGA BIOS fnct. 0x0E: teletype
.loop:   
    lodsb             ; grab a byte from SI
    test al, al       ; NUL?
    jz .done          ; if the result is zero, get out
    int 0x10          ; otherwise, print out the character!
    jmp .loop
.done:
    ret

empty_8042:
    call Waitingloop
    in al, 0x64
    cmp al, 0xff      ; ... no real kbc at all?
    je .done
   
    test al, 1        ; something in input buffer?
    jz .no_output
    call Waitingloop
    in al, 0x60       ; yes: read buffer
    jmp empty_8042    ; and try again
   
.no_output:
    test al, 2        ; command buffer empty?
    jnz empty_8042    ; no: we can't send anything new till it's empty
.done:
    ret

Waitingloop:                   
    mov cx,0xFFFF
.loop_start:
    dec cx     
    jnz .loop_start
    ret       

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;    read sectors from floppy disk to buffer ES:BX in memory    ;
;                                                               ; 
;    input:                                                     ;
;    di         - number of sectors                             ;
;    bx         - segment                                       ;
;    ch, dh, cl - cylinder, head, sector                        ; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

read_sectors:

.reset_floppy:
    xor ax, ax        ; mov ax, 0  => function "reset"
    int 0x13           
    jc .reset_floppy  ; trouble? try again 
	
.next:
    mov si, progressmsg   ; show progress
    call print_string
	mov es, bx
    xor bx, bx

.again:
    mov dl, [bootdrive]
    mov ax, 0x0201
    int 0x13
    jc .again

    inc cl
    cmp cl, 19
    jl .loop
    mov cl, 1

    inc dh
    cmp dh, 2
    jl .loop
    mov dh, 0

    inc ch
    cmp ch, 80
    jae .error

.loop:
	mov bx, es
    add bx, 512/16
    sub di, 1
	jnz .next

.done:
    mov dx,0x3F2      ; switch off floppy disk motor
    mov al,0x0C
    out dx,al 
    ret	

.error:
    mov si, errormsg   ; show error message
    call print_string
.end:
    jmp .end	
 	
;;;;;;;;;;;;
; Includes ;
;;;;;;;;;;;;

%include "gdt.inc"
	
;;;;;;;;
; data ;
;;;;;;;;

    bootdrive db 0    
    loadmsg db "bootloader message: loading kernel ...",13,10,0
    errormsg db "bootloader message: sector read error ...",13,10,0
	progressmsg db "*",0
	
    times 510-($-$$) hlt
    db 0x55
    db 0xAA

