#include "os.h"
#include "kheap.h"
#include "task.h"
#include "initrd.h"
#include "syscall.h"
#include "shared_pages.h"

extern ULONG file_data_start;
extern ULONG file_data_end;
ULONG address_user;
UCHAR address_TEST[4096];
UCHAR buf[4096];
UCHAR flag1 = 0; // status of user-space-program

extern ULONG placement_address;
heap_t* kheap = 0;

static void init()
{
    kheap = 0;
    page_directory_t* kernel_directory  = 0;
    page_directory_t* current_directory = 0;
    placement_address = 0x200000;
}

int main()
{
    init();
    k_clear_screen(); settextcolor(14,0);
    printformat("PrettyOS [Version 0.1.0085]  (C) 2009 henkessoft.de\n\n");
    gdt_install();
    idt_install();
    isrs_install();
    irq_install();
    initODA();
    timer_install();
    keyboard_install();
    paging_install();

    kheap = create_heap(KHEAP_START, KHEAP_START+KHEAP_INITIAL_SIZE, KHEAP_MAX, 1, 0); // SV and RW

    ///
    #ifdef _DIAGNOSIS_
    settextcolor(2,0);
    printformat("rd_start: ");
    settextcolor(15,0);
    #endif
    ///

    settextcolor(15,0);
    ULONG ramdisk_start = k_malloc(0x30000, 0, 0); // Ram Disk at the kernel heap
    //fs_root = install_initrd(ramdisk_start); // implement Ram Disk into VFS /// ASSERT at kheap
    tasking_install(); // ends with sti()

    // program provided by data.asm
    k_memcpy((void*)address_TEST, &file_data_start, (ULONG)&file_data_end - (ULONG)&file_data_start);

    /// peeking data from elf exec format ///
    ULONG elf_vaddr     = *( (ULONG*)( address_TEST + 0x3C ) );
    ULONG elf_offset    = *( (ULONG*)( address_TEST + 0x38 ) );
    ULONG elf_filesize  = *( (ULONG*)( address_TEST + 0x44 ) );

    ///
    #ifdef _DIAGNOSIS_
    settextcolor(5,0);
    printformat("virtual address: %x offset: %x size: %x", elf_vaddr, elf_offset, elf_filesize);
    printformat("\n\n");
    settextcolor(15,0);
    #endif
    ///

    if( (elf_vaddr == 0x400000) || (elf_offset<0x130) || (elf_filesize<0x1000) )
    {
        flag1=1;
    }

    // address_TEST (ring 0) ==> address_user (ring 3)
    if(flag1==1)
    {
        address_user = elf_vaddr;
        k_memcpy((void*)address_user, address_TEST + elf_offset, elf_filesize); // ELF LOAD: Offset VADDR FileSize
    }

    pODA->ts_flag = 1; // enable task_switching

    if(flag1==1)
    {
        create_task ((void*)getAddressUserProgram(),3); // program in user space (ring 3) takes over
    }
    else
    {
        settextcolor(4,0);
        printformat("Program not found.\n");
        settextcolor(15,0);

    }

    while(TRUE){/*  */}
    return 0;
}
